@@ -34,7 +34,9 @@
 #include "data/reconcile.h"
 #include "data/write.h"
 
+#ifdef CONFIG_DEBUG_FS
 #include "debug/async_objs.h"
+#endif
 #include "debug/debug.h"
 #include "debug/sysfs.h"
 
@@ -442,18 +444,7 @@ static void bch2_fs_read_only_async(struct bch_fs *c)
 	queue_work(system_long_wq, &c->read_only_work);
 }
 
-bool bch2_fs_emergency_read_only(struct bch_fs *c)
-{
-	bool ret = !test_and_set_bit(BCH_FS_emergency_ro, &c->flags);
-
-	bch2_journal_halt(&c->journal);
-	bch2_fs_read_only_async(c);
-
-	wake_up(&bch2_read_only_wait);
-	return ret;
-}
-
-static bool __bch2_fs_emergency_read_only2(struct bch_fs *c, struct printbuf *out, bool locked)
+static bool __bch2_fs_emergency_read_only(struct bch_fs *c, struct printbuf *out, bool locked)
 {
 	bool ret = !test_and_set_bit(BCH_FS_emergency_ro, &c->flags);
 
@@ -474,20 +465,14 @@ static bool __bch2_fs_emergency_read_only2(struct bch_fs *c, struct printbuf *ou
 }
 
 /* Returns true if going ERO, false if we already are */
-bool bch2_fs_emergency_read_only2(struct bch_fs *c, struct printbuf *out)
+bool bch2_fs_emergency_read_only(struct bch_fs *c, struct printbuf *out)
 {
-	return __bch2_fs_emergency_read_only2(c, out, false);
+	return __bch2_fs_emergency_read_only(c, out, false);
 }
 
-bool bch2_fs_emergency_read_only_locked(struct bch_fs *c)
+bool bch2_fs_emergency_read_only_locked(struct bch_fs *c, struct printbuf *out)
 {
-	bool ret = !test_and_set_bit(BCH_FS_emergency_ro, &c->flags);
-
-	bch2_journal_halt_locked(&c->journal);
-	bch2_fs_read_only_async(c);
-
-	wake_up(&bch2_read_only_wait);
-	return ret;
+	return __bch2_fs_emergency_read_only(c, out, true);
 }
 
 static int __bch2_fs_read_write(struct bch_fs *c, bool early)
@@ -626,7 +611,10 @@ static void __bch2_fs_free(struct bch_fs *c)
 	bch2_fs_buckets_waiting_for_journal_exit(c);
 	bch2_fs_btree_exit(c);
 	bch2_fs_accounting_exit(c);
+
+#ifdef CONFIG_DEBUG_FS
 	bch2_fs_async_obj_exit(c);
+#endif
 
 	enumerated_ref_exit(&c->writes);
 	kfree(rcu_dereference_protected(c->disk_groups, 1));
@@ -825,21 +813,21 @@ static bool check_version_upgrade(struct bch_fs *c, struct printbuf *out)
 
 	if (new_version > old_version) {
 		if (old_version < bcachefs_metadata_required_upgrade_below)
-			prt_str_indented(out, "Version upgrade required:\n");
+			prt_str(out, "Version upgrade required:\n");
 
 		if (old_version != c->sb.version) {
-			prt_str_indented(out, "Version upgrade from ");
+			prt_str(out, "Version upgrade from ");
 			bch2_version_to_text(out, c->sb.version_upgrade_complete);
-			prt_str_indented(out, " to ");
+			prt_str(out, " to ");
 			bch2_version_to_text(out, c->sb.version);
-			prt_str_indented(out, " incomplete\n");
+			prt_str(out, " incomplete\n");
 		}
 
 		prt_printf(out, "Doing %s version upgrade from ",
 			   BCH_VERSION_MAJOR(old_version) != BCH_VERSION_MAJOR(new_version)
 			   ? "incompatible" : "compatible");
 		bch2_version_to_text(out, old_version);
-		prt_str_indented(out, " to ");
+		prt_str(out, " to ");
 		bch2_version_to_text(out, new_version);
 		prt_newline(out);
 
@@ -849,7 +837,7 @@ static bool check_version_upgrade(struct bch_fs *c, struct printbuf *out)
 		passes = ext->recovery_passes_required[0] & ~passes;
 
 		if (passes) {
-			prt_str_indented(out, "Upgrade requires recovery passes: ");
+			prt_str(out, "Upgrade requires recovery passes: ");
 			prt_bitflags(out, bch2_recovery_passes,
 				     bch2_recovery_passes_from_stable(le64_to_cpu(passes)));
 			prt_newline(out);
@@ -860,9 +848,9 @@ static bool check_version_upgrade(struct bch_fs *c, struct printbuf *out)
 
 	if (new_version > c->sb.version_incompat_allowed &&
 	    c->opts.version_upgrade == BCH_VERSION_UPGRADE_incompatible) {
-		prt_str_indented(out, "Now allowing incompatible features up to ");
+		prt_str(out, "Now allowing incompatible features up to ");
 		bch2_version_to_text(out, new_version);
-		prt_str_indented(out, ", previously allowed up to ");
+		prt_str(out, ", previously allowed up to ");
 		bch2_version_to_text(out, c->sb.version_incompat_allowed);
 		prt_newline(out);
 
@@ -896,7 +884,7 @@ static int bch2_fs_opt_version_init(struct bch_fs *c, struct printbuf *out)
 	    bch2_fs_will_resize_on_mount(c))
 		set_bit(BCH_FS_may_upgrade_downgrade, &c->flags);
 
-	prt_str_indented(out, "starting version ");
+	prt_str(out, "starting version ");
 	bch2_version_to_text(out, c->sb.version);
 	prt_newline(out);
 
@@ -911,7 +899,7 @@ static int bch2_fs_opt_version_init(struct bch_fs *c, struct printbuf *out)
 		if (v == bch2_opt_get_by_id(&bch2_opts_default, i))
 			continue;
 
-		prt_str_indented(out, first ? "with options: " : ",");
+		prt_str(out, first ? "with options: " : ",");
 		first = false;
 		bch2_opt_to_text(out, c, c->disk_sb.sb, opt, v, OPT_SHOW_MOUNT_STYLE);
 	}
@@ -962,21 +950,21 @@ static int bch2_fs_opt_version_init(struct bch_fs *c, struct printbuf *out)
 
 		u64 sb_passes = bch2_recovery_passes_from_stable(le64_to_cpu(ext->recovery_passes_required[0]));
 		if (sb_passes) {
-			prt_str_indented(out, "superblock requires following recovery passes to be run: ");
+			prt_str(out, "superblock requires following recovery passes to be run: ");
 			prt_bitflags(out, bch2_recovery_passes, sb_passes);
 			prt_newline(out);
 		}
 
 		u64 btrees_lost_data = le64_to_cpu(ext->btrees_lost_data);
 		if (btrees_lost_data) {
-			prt_str_indented(out, "superblock indicates damage to following btrees:  ");
+			prt_str(out, "superblock indicates damage to following btrees:  ");
 			prt_bitflags(out, __bch2_btree_ids, btrees_lost_data);
 			prt_newline(out);
 		}
 
 		if (test_bit(BCH_FS_may_upgrade_downgrade, &c->flags)) {
 			if (bch2_check_version_downgrade(c)) {
-				prt_str_indented(out, "Version downgrade required");
+				prt_str(out, "Version downgrade required");
 
 				__le64 passes = ext->recovery_passes_required[0];
 				bch2_sb_set_downgrade(c,
@@ -984,7 +972,7 @@ static int bch2_fs_opt_version_init(struct bch_fs *c, struct printbuf *out)
 						      BCH_VERSION_MINOR(c->sb.version));
 				passes = ext->recovery_passes_required[0] & ~passes;
 				if (passes) {
-					prt_str_indented(out, ", running recovery passes: ");
+					prt_str(out, ", running recovery passes: ");
 					prt_bitflags(out, bch2_recovery_passes,
 						     bch2_recovery_passes_from_stable(le64_to_cpu(passes)));
 				}
@@ -1018,13 +1006,13 @@ static int bch2_fs_opt_version_init(struct bch_fs *c, struct printbuf *out)
 
 	if (BCH_SB_INITIALIZED(c->disk_sb.sb)) {
 		if (!(c->sb.features & BIT_ULL(BCH_FEATURE_new_extent_overwrite))) {
-			prt_str_indented(out, "feature new_extent_overwrite not set, filesystem no longer supported\n");
+			prt_str(out, "feature new_extent_overwrite not set, filesystem no longer supported\n");
 			return -EINVAL;
 		}
 
 		if (c->sb.version_min < bcachefs_metadata_version_btree_ptr_sectors_written) {
-			prt_str_indented(out, "version_min < version_btree_ptr_sectors_written\n");
-			prt_str_indented(out, "filesystem needs upgrade from older version; run fsck from older bcachefs-tools to fix\n");
+			prt_str(out, "version_min < version_btree_ptr_sectors_written\n");
+			prt_str(out, "filesystem needs upgrade from older version; run fsck from older bcachefs-tools to fix\n");
 			return -EINVAL;
 		}
 	}
@@ -1143,7 +1131,9 @@ static int bch2_fs_init(struct bch_fs *c, struct bch_sb *sb,
 		return bch_err_throw(c, ENOMEM_fs_other_alloc);
 
 	try(bch2_blacklist_table_initialize(c));
+#ifdef CONFIG_DEBUG_FS
 	try(bch2_fs_async_obj_init(c));
+#endif
 	try(bch2_fs_btree_init(c));
 	try(bch2_fs_buckets_waiting_for_journal_init(c));
 	try(bch2_fs_compress_init(c));
@@ -1201,11 +1191,13 @@ static int bch2_fs_init(struct bch_fs *c, struct bch_sb *sb,
 		darray_for_each(*sbs, sb)
 			try(bch2_dev_attach_bdev(c, sb, out));
 
-	/*
-	 * Do this early, so that we never expose a filesystem object that
-	 * hasn't been version downgraded
-	 */
-	try(bch2_fs_opt_version_init(c, out));
+	if (!c->opts.no_version_check) {
+		/*
+		 * Do this early, so that we never expose a filesystem object that
+		 * hasn't been version downgraded
+		 */
+		try(bch2_fs_opt_version_init(c, out));
+	}
 
 	/*
 	 * just make sure this is always allocated if we might need it - mount
@@ -1241,6 +1233,11 @@ static int bch2_fs_may_start(struct bch_fs *c, struct printbuf *err)
 {
 	unsigned flags = 0;
 
+	if (c->opts.no_version_check) {
+		prt_printf(err, "Cannot start with opts.no_version_check\n");
+		return -EINVAL;
+	}
+
 	switch (c->opts.degraded) {
 	case BCH_DEGRADED_very:
 		flags |= BCH_FORCE_IF_DEGRADED|BCH_FORCE_IF_LOST;
@@ -1464,7 +1461,7 @@ static struct bch_fs *__bch2_fs_open(darray_const_str *devices,
 	if (ret)
 		goto err;
 
-	{
+	if (out->buf) {
 		/* Log opt_version_init() message before doing actual filesystem startup */
 		CLASS(printbuf, msg_with_prefix)();
 		prt_str(&msg_with_prefix, out->buf);
