@@ -91,6 +91,7 @@ struct printbuf {
 	enum printbuf_si	si_units:1;
 	bool			human_readable_units:1;
 	bool			has_indent_or_tabstops:1;
+	bool			may_vmalloc:1;
 	u8			nr_tabstops;
 
 	/*
@@ -235,11 +236,6 @@ static inline void prt_bytes(struct printbuf *out, const void *b, unsigned n)
 }
 
 static inline void prt_str(struct printbuf *out, const char *str)
-{
-	prt_bytes(out, str, strlen(str));
-}
-
-static inline void prt_str_indented(struct printbuf *out, const char *str)
 {
 	bch2_prt_bytes_indented(out, str, strlen(str));
 }
